/* Content:
   - aunfold.Hypergraph
   - aunfold.Label
   - aunfold.MatchHandler
*/


#ifndef HYPERGRAPH_H
#define HYPERGRAPH_H

#include <string>
#include <iostream>
#include <list>
#include <map>
#include <set>
#include <stack>

#include "Globals.h"

namespace aunfold {



  /** Type definition for edge labels. Make shure operator== and operator!=
   *  are defined for LABELTYPE.*/
  typedef std::string LABELTYPE;
  

  /** forward declaration (see below)*/
  class MatchHandler;

  /** Representation of a Hypergraph */

  class Hypergraph
    {
    public:
      Hypergraph();
      Hypergraph(const Hypergraph& h); // copy constructor
      ~Hypergraph();

      /** Adds a new vertex to the hypergraph. The ID of the created
	  vertex is returned. */
      IDTYPE addVertex();

      /** Adds a new vertex to the hypergraph. The passed ID must be unique.
          If id is already a vertex of the graph, nothing is being done*/
      IDTYPE addVertex(IDTYPE id);
      /** Adds a new hyperedge to the hypergraph.
          connection is a list of vertices in the hypergraph.
	  label is a label for the new edge.
	  The ID of the new edge is returned.
      */
      IDTYPE addEdge(std::list<IDTYPE> connection, LABELTYPE label);

      /** Adds a new hyperedge to the hypergraph. the passed ID must be unique
          If id is already an edge of the graph, nothing is being done*/
	  
      IDTYPE addEdge(std::list<IDTYPE> connection, LABELTYPE label, IDTYPE id);


      
      /** find a subgraph matching of h in *this.
	  Result is a mapping from (vertices UNION edged) of h to
          (vertices UNION edges) of this. The mapping is defined for all
	  edges and for those vertices connected to some edge in h (i.e.
	  the mapping is total if h is a left hand side of a simple 
	  transformation rule).
	  Everytime a matching is found, handler.useThisMatch() is called.
	  If useThisMatch returns true, findMatching returns that matching.
	  Otherwise the search continues. If no matching was found,
	  an empty map is returned.
      */
      std::map<IDTYPE,IDTYPE> findMatching(Hypergraph& h, MatchHandler& handler, long depth, long unfdepth);

      
      /** Merges h into this. PRE: vertices and edges of h and this are
          disjunct.  */

      void uniongraphs(const Hypergraph& h);
      

      /** returns the Label of edge e) */
      LABELTYPE getLabel(IDTYPE edge) const;

      /** returns the set of vertices */
      std::set<IDTYPE> getVertices();

      /** returns the set of edges */
      std::set<IDTYPE> getEdges();
      /** returns the connected vertices of edge e */
      std::list<IDTYPE> getConnectedVertices(IDTYPE edge) const;
      
      /** returns the connected edges of vertex v */
      std::multiset<IDTYPE> getConnectedEdges(IDTYPE vertex) const;
      friend std::ostream& operator<<(std::ostream& os, Hypergraph& hg);

      // depth
      bool setDepth(IDTYPE id, long depth, VERTEDGE v);
      bool getDepth(IDTYPE id, long * pdepth, VERTEDGE v);

      void initDepth();

      std::map<IDTYPE, long> getDepthTable(VERTEDGE v);
      void setDepthTable(std::map<IDTYPE, long> & depthtable, VERTEDGE v);

      void setIdLabel(IDTYPE id, std::string ls);
      std::string getIdLabel(IDTYPE id);

    private:
      // backtracking function to find a subgraph matching
      // returns true if a matching was found
      bool findNextMatching(Hypergraph &h,
			    std::stack<std::set<IDTYPE>::iterator> &mappings,
			    std::stack<IDTYPE> &unmapped_edges,
			    std::map<IDTYPE,IDTYPE> &phi_v,
			    std::map<IDTYPE,IDTYPE> &phi_e,
			    MatchHandler& matchhandler,
			    long depth,
			    long unfdepth);
			    


      
      // Hypergraph datastructure:
      std::set<IDTYPE> vertices;
      std::set<IDTYPE> edges;
      
      std::map<IDTYPE,std::list<IDTYPE> > connectedV; // Vertices connected to an edge
      std::map<IDTYPE,std::multiset<IDTYPE> > connectedE; // reverse mapping of connectedV
      
      std::map<IDTYPE,LABELTYPE> labels;  // Labeling function

      // original labels for ids in hypergraph
      // is important for startgraph and rules
      // can be empty otherwise
      std::map<IDTYPE, std::string> idLabels; 

      std::map<IDTYPE, long> depthmapEdges;

      std::map<IDTYPE, long> depthmapVertices;   
};




  
  /** ABC for Labels to be attached to edges of a Hypergraph 
class Label
{
 public:
  virtual bool equals(const Label& l) = 0;

  }; */


/** ABC for handlers for found matches */
class MatchHandler
{
 public:
  virtual bool useThisMatch(std::map<IDTYPE,IDTYPE> match) = 0;
};


}

#endif



