#ifndef PTNET_H
#define PTNET_H

#include <assert.h>
#include <list>
#include <set>
#include <map>
#include <limits.h>

#include "Globals.h"

using namespace std;

namespace aunfold
{
  /**
   * Definition of a marking of a Place/Transition net.
   * A marking must be a total mapping from the places
   * to unsigned integers; insert p->0 mappings where
   * necessary
   */

  typedef map<IDTYPE,unsigned> PTMarking;

  // "infinity" symbol; define as largest possible unsigned value
  const unsigned OMEGA = UINT_MAX;


  /** Transition of a PTNet */
  class PTNetTransition
    {
    public:
      PTNetTransition(set<IDTYPE> p_preset,
                      set<IDTYPE> p_postset,
                      map<IDTYPE,unsigned>p_preweight,
                      map<IDTYPE,unsigned>p_postweight,
                      IDTYPE p_associatedRule);

      PTNetTransition(const PTNetTransition& t);

      bool isEnabledBy(PTMarking marking) const;
      
      PTMarking fire(const PTMarking& marking) const;

      set<IDTYPE> getPreset() const;
      set<IDTYPE> getPostset() const;
      unsigned getPreWeight(IDTYPE place) const;
      unsigned getPostWeight(IDTYPE place) const;
      IDTYPE getID() const;

      /**
       * ID of the transformation rule associated to the
       * transition, as returned by the labeling function p
       */
      IDTYPE getAssociatedRule() const;

      friend ostream& operator<<(ostream& os, PTNetTransition& trans);
      
      
    private:
      set<IDTYPE> preset;
      set<IDTYPE> postset;
      map<IDTYPE,unsigned> preweight; //Weight of the arcs from place IDTYPE
      map<IDTYPE,unsigned> postweight; //Weight of the arcs to place IDTYPE
      IDTYPE id;
      IDTYPE associatedRule;
      
    };
    
  class PTNet;
    
  class CovGraph
  {
    public:
      void addNode(const PTMarking& marking);
      void addEdge(const PTMarking& m1, const PTMarking& m2, const PTNetTransition& transition);
      void clear();
      bool containsNode(const PTMarking& marking);  
      bool getNodeWithSmallerMarking(const PTMarking& node, const PTMarking& marking, PTMarking& smallerNode);
    
    private:
      map<PTMarking,IDTYPE> m_nodeMapping;
      map<IDTYPE,PTMarking> m_markingMapping;

      /**
       * 1. IDTYPE: Node of the coverability graph
       * 2. IDTYPE: Successor node in coverability graph
       * 3. IDTYPE: ID of the corresponding transition in the PTNet
       */
      map<IDTYPE,map<IDTYPE,IDTYPE> > m_succ;
      map<IDTYPE,map<IDTYPE,IDTYPE> > m_pred;

      list<pair<IDTYPE,list<PTNetTransition> > > m_work;
      
    friend ostream& operator<<(ostream& os, CovGraph& cg);
    friend class PTNet;
  };
      
  /**
   * Definition of a Place/Transition net
   */
  class PTNet
    {
    public:
      PTNet(set<IDTYPE> p_places,
      list<PTNetTransition> p_transitions,
      PTMarking p_initial_marking);
      PTNet(const PTNet& net);

      PTMarking getInitialMarking() const
      {
        return initial_marking;
      };
      
      list<PTNetTransition> getEnabledTransitions(PTMarking marking) const;
      
      bool isCoverableMarking(const PTMarking& marking);

      PTNetTransition& getTransition(IDTYPE id);

      /**
       * returns true iff element1 <N element2 where <N is the causality
       * relation as defined in definition 5
       */
      bool inCausalityRelation(IDTYPE element1, IDTYPE element2);

      set<IDTYPE> getPlaces();
      list<PTNetTransition> getTransitions();

      void addPlace(IDTYPE p);
      void addTransition(PTNetTransition t);
      
      friend ostream& operator<<(ostream& os, PTNet& ptn);
  /** No descriptions */
  void deleteTransition(IDTYPE id);

    private:
      set<IDTYPE> places;
      list<PTNetTransition> transitions;
      
      CovGraph m_covGraph;

      PTMarking initial_marking;

      // helper functions for inCausalityRelation()
      bool inCausalityRelationPlace(IDTYPE element1,
                                    IDTYPE element2,
                                    set<IDTYPE>& visited);
      bool inCausalityRelationTransition(IDTYPE element1,
                                         IDTYPE element2,
                                         set<IDTYPE>& visited);

    };
    
    
    

  /**
   * helper functions for calculation of the coverability graph
   */
  bool isEqual(const PTMarking& m1, const PTMarking& m2);

  /**
   * returns true iff m1 is smaller than m2 or equal to m2
   */
  bool isLower(const PTMarking& m1, const PTMarking& m2);

  bool isStrictlyLower(const PTMarking& m1, const PTMarking& m2);

      
}

#endif
