#include "InputCvs.h"
#include <fstream>
//#include <cstdlib>
#include <sstream>

InputCvs::InputCvs() { // constructor
  final = new ConstraintSet;
  initial = new ConstraintSet;
  ptSet = new PredicateTransformerSet;
  placeNames = new stringMap;
  transformerNames = new stringMap;
  verbose = false;
}

// methods, for obtaining parsed infos
PredicateTransformerSet* InputCvs::getPredicateTransformerSet() { return ptSet; }
ConstraintSet* InputCvs::getFinalConstraintSet() { return final; }  
ConstraintSet* InputCvs::getInitialConstraintSet() { return initial; }  
stringMap* InputCvs::getTransformerNames() { return transformerNames; }  
stringMap* InputCvs::getPlaceNames() { return placeNames; }  

// exit with error message
void InputCvs::error_exit(const string in) {
  cout << in << endl; 
  std::exit(1);
}

/* read the contents of a text file,
   and return content as a string,
   skipping blanks.
*/
string InputCvs::readFile(const char* file) { 
  ostringstream os;
  char ch;
  std::ifstream input(file);
  if (!input) { 
    this->error_exit("Failed to open File: \"" +string(file)+"\"");
  }
  
  while (input.get(ch)) { 
    if ((ch != ' ') && (ch != '\n')) {
      os.put(ch);
    }
  }
  return os.str();
}

/*
  parse input files for information in LoLA-format.
  file1 has places, initial marking and Transition net,
  file2 has end/test marking/constraint.
  If any error occures, return false,
  else return true.
*/
bool InputCvs::run(const char* file1, const char* file2) { 
  string s; // for saving raw content of files
  string::size_type a,b,c; // needed for String-Parsing
  string token = "PLACE";
  string error_str = ""; // used to store error messages
  
  s = this->readFile(file1); // read content of file1
  
  error_str = "Error while parsing Places in File \""+ string(file1) + "\":\n";
  
  // get the place names
  token = "PLACE";
  a = s.find(token);
  if(a < s.size()) {
    b = s.find(";");
    if(b < s.size()) {
      this->setPlaces(string(s,a,(b-a)+1));
      s.erase(0,b+1);
    } else {
      this->error_exit(error_str + "No delimiting \";\" found");
      return false;
    }
  } else {
    this->error_exit(error_str + "Token \"PLACE\" not found.");
    return false;
  }
  
  error_str = "Error while parsing Marking in File \""+ string(file1) + "\":\n";
  
  // get Initial Constraints
  for (a = s.find("MARKING"); a < s.size(); a = s.find("MARKING")) {
    b = s.find(";");
    if (b < s.size()) {
      this->addConstraint(initial,string(s,a,(b-a)+1));
      s.erase(0,b+1);
    } else {
      this->error_exit(error_str + "No delimiting \";\" found");
      return false;
    }
  }
  
  error_str = "Error while parsing Transition in File \""+ string(file1) + "\":\n";
  
  // parse Transitions
  token = "TRANSITION";
  a = s.find(token);
  while((!(s.empty())) && a < s.size()) {
    b = s.find(";");
    c = s.find(";",b+1);
    if((b > a) && (c > a) && (c < s.size()) && (b < c)) {
      this->addPredicateTransformer(string(s,a,(c-a)+1));
      s.erase(0,c+1);
    } else {
      this->error_exit(error_str + "No matching \";\" found");
      return false;
    }
    a = s.find(token);
  }
  
  error_str = "Error while parsing Marking in File \""+ string(file1) + "\":\n";
  
  // get Final Constraints from file2
  s = this->readFile(file2);
  
  for (a = s.find("MARKING"); a < s.size(); a = s.find("MARKING")) {
    b = s.find(";");
    if (b < s.size()) {
      this->addConstraint(final,string(s,a,(b-a)+1));
      s.erase(0,b+1);
    } else {
      this->error_exit(error_str + "No delimiting \";\" found");
      return false;
    }
  }
  
  if (verbose) { cout << this->asString(); }
  return true; 
}

/* get a marking in LoLA-format as input-string,
   parse it, and save it in the Constraint Set.
   While parsing, check that place names are known.
*/
void InputCvs::addConstraint(ConstraintSet* ctSet,string in) {
  string const token = "MARKING";
  int resCode = 0;
  string error_str = "Error while parsing Marking:\n";
  Constraint* newCT = new Constraint; // save the marking
  string::size_type a,b,c;
  string place,val; // a marking is a set of place-value-pairs ...
  int place_int, val_int; 

  in.erase(0,token.size());

  c = in.find_first_of(",;");
  while((!(in.empty())) && c < in.size()) {
    
    b = in.find(":"); // find delimiter between place and value
    
    if (b < c) {
      place = string(in,0,b);
      val = string(in,b+1,c-b);
      
      // check if place-name is not empty
      if (!place.empty()) {
	
	// is the place name known?
	place_int = placeNames->has(place);
	
	if (place_int == 0) { // error, place name unknown
	  this->error_exit(error_str+
			   "Marking refers to unknown Place \""+
			   place +"\"\n");
	  
	} else { // place name known, parse Integer value
	  resCode = sscanf(val.c_str(),"%d",&val_int);
	  if (resCode == 0) { // error, could not parse Integer
	    this->error_exit(error_str+
			     "no Integer value at place \""+
			     place +"\"\n");
	  } else {
	    newCT->set(place_int,val_int); // update the Constraint
	  }
	}
      } else {
	this->error_exit(error_str+"Marking includes empty place name.\n");
      }
    }
    in.erase(0,c+1); // erase sub-string until one of [,;]
    c = in.find_first_of(",;"); // find next delimiter
  }
  // marking is parsed, add constraint to constraint set
  ctSet->add(newCT);
}


/* get a Transition in LoLA-format as input-string,
   parse it, and save it in the Transition Set.
   While parsing, check if place names are known.
   If any error occurs, return false.
*/
bool InputCvs::addPredicateTransformer(string in) {  
  // save parsed place-value-pairs
  PredicateTransformer* newPT = new PredicateTransformer;

  string::size_type a,b,c;

  // store transtion- and place-name, and value-string
  string place,val,trans; 

  // store transtion ID, place ID and value
  int place_int, val_int,transition_int;

  int res_code = 0;
  string error_str = ""; // used for error messages;
  string token = "TRANSITION";

  error_str = "Error while parsing Transition:\n";   
  
  // parse Transition token
  b = in.find(token);
  if (b > in.size()) { // "TRANSITION" not found
    this->error_exit(error_str+"Token \"TRANSITTION\" not found.");
    return false;
  }
  in.erase(0,b+token.size());
  
  // parse Transition name
  token = "CONSUME";
  b = in.find(token);
  if (b > in.size()) { // name delimter not found
    this->error_exit(error_str+"Token \"CONSUME\" not found.");
    return false;
  }

  trans = string(in,0,b); // name of Transition
  if (trans.empty()) {
    this->error_exit(error_str+"Empty Transition name found.");
    return false;
  }
  
  error_str = "Error while parsing Transition \""+trans+"\":\n";
  
  // check if Transition Name already available
  transition_int = transformerNames->has(trans); 
  if (transition_int > 0) {
    this->error_exit(error_str + "Transition Name already available.");
    return false;
  }
  in.erase(0,b+token.size());
  // end: parse Transition name
  
  // set name
  newPT->setName(trans);
  
  // begin: parse CONSUME-Funktion
  token = "PRODUCE";
  a = in.find(token); 
  if (a >= in.size()) { // delimiter not found
    this->error_exit(error_str+"Token \"PRODUCE\" not found.");
    return false;
  }
  
  c = in.find_first_of(",;");  // find token delimter for place-value-pair
  while(c < a) { // "," or ";" found before "PRODUDE"
    b = in.find(":");
    if (b < c) {  // ":" found, check for place name and value
      place = string(in,0,b); // place name
      if (place.empty()) { 
	this->error_exit(error_str + "Empty place name found.");
	return false;
      }
      place_int = placeNames->has(place);
      if (place_int == 0) {
	this->error_exit(error_str + 
			 "Transition refers to unknown place \"" + place + "\".");
	return false;
      } else { // place name is OK, check its value
	val = string(in,b+1,c-b); // place value
	res_code = sscanf(val.c_str(),"%d",&val_int);
	if (res_code == 0) {
	  this->error_exit(error_str + "Value at place \"" +
			   place + "\" is not an Integer value.");
	  return false;
	} else {
	  newPT->setIn(place_int,val_int);
	}
      }
    }
    in.erase(0,c+1);
    a = in.find(token); 
    c = in.find_first_of(",;");  // find token delimter for place-value-pair
  }
  
  // end: parse CONSUME-Funktion

  token = "PRODUCE";
  a = in.find(token); 
  in.erase(0,a+token.size()); // remove "PRODUCE" token
  
  // begin: parse PRODUCE-Funktion
  c = in.find_first_of(",;");  // find token delimter for place-value-pair
  while((!(in.empty())) && c < in.size()) { // "," or ";" found
    b = in.find(":");
    if (b < c) {  // ":" found, check for place name and value
      place = string(in,0,b); // place name
      if (place.empty()) { 
	this->error_exit(error_str + "Empty place name found.");
	return false;
      }
      place_int = placeNames->has(place);
      if (place_int == 0) {
	this->error_exit(error_str + 
			 "Transition refers to unknown place \"" + place + "\".");
	return false;
      } else { // place name is OK, check its value
	val = string(in,b+1,c-b); // place value
	res_code = sscanf(val.c_str(),"%d",&val_int);
	if (res_code == 0) {
	  this->error_exit(error_str + "Value at place \"" +
			   place + "\" is not an Integer value.");
	  return false;
	} else {
	  newPT->setOut(place_int,val_int);
	}
      }
    }
    in.erase(0,c+1);
    c = in.find_first_of(",;");  // find token delimter for place-value-pair
  }
  // end: parse PRODUCE-Funktion
  
  // get ID of new Transformer after adding to Transformer Set,
  // and save this ID in Transformer Names
  if (newPT->empty()) {
    this->error_exit(error_str + "Transition is empty.");
    return false;
  } else {
    transformerNames->set(trans,ptSet->add(newPT));
    return true;
  }
  return false;
}

// parse and save Place names from input string,
// empty place names are skipped.
void InputCvs::setPlaces(string in) {
  const string token = "PLACE";
  string::size_type c;
  string place;
  in.erase(0,token.size());
  c = in.find_first_of(",;");
  
  while(c < in.size()) {
    place = string(in,0,c);
    if (!(place.empty())) { 
      placeNames->add(place);      
    }
    in.erase(0,c+1);
    c = in.find_first_of(",;");
  }
}

// return parsed informations as string
string InputCvs::asString() {
  ostringstream os;
  os << "Initial Markings:\n"
     << this->getInitialConstraintSet()->asString(placeNames)
     << "Test Markings:\n"
     << this->getFinalConstraintSet()->asString(placeNames)
     << "Net:\n"
     << this->getPredicateTransformerSet()->asString(placeNames,transformerNames);
  return os.str();
}
