#include "CG.h"
#include <time.h>
#include <sstream>

// constructor, initialize internal variables using input
CG::CG(TransitionSet* T,Node* initialMarking,
       stringMap* inPlaceNames,stringMap* inTransNames) { 
  
  timer = double(clock()) / CLOCKS_PER_SEC; // reset timer
  
  verb = false;
  quiet = false;
  
  transitions = T;
  nodes = new NodeSet;
  
  // add the initial Marking to the Set of Nodes
  nodes->add(initialMarking);
  
  // save reference to names of places and transitions
  placeNames = inPlaceNames;
  transitionNames = inTransNames;
}


void CG::setVerbose() { verb = true; } // print more messages to console

void CG::setQuiet() { quiet = true; } // print minimum of messages to console

// return processed time as string, rounded to two places after decimal point.
string CG::getTime() { 
  ostringstream os;
  os.flags(ios_base::fixed);
  os.precision(2);
  if (timer < double(0)) { 
    timer = 0.0;
  }
  os << ((double(clock()) / CLOCKS_PER_SEC) - timer);
  return os.str();
}

// return names of known places
stringMap* CG::getPlaceNames() { return placeNames; }

// return names of known transitions
stringMap* CG::getTransitionNames() { return transitionNames; }

// return reference to transition set
TransitionSet* CG::getTransitionSet() {   return transitions; }


// is this CG empty, ie. does it have any nodes?
bool CG::empty() { return nodes->isEmpty(); }

// retrun string-representation of this CG
string CG::asString() {
  ostringstream os;
  os << "Coverability Graph\n"
     << nodes->asString(placeNames);
  return os.str();
}

// central method: run the CG-algorithm
bool CG::run() {
  intSet* unvisitedNodes = new intSet; // which nodes are unvisited yet?
  intSet* visitedNodes = new intSet; // which nodes are visited

  Node* currNode = new Node;
  Node* newNode = new Node;
  Transition* currTrans = new Transition;

  intSet* preNodes = new intSet; // save predecessor nodes IDs
  
  int id = 0; // ID of visited node
  int newId = 0; // ID of new generated node
  int k = 0;
  
  // get all IDs of node set, ie. all nodes are marked to be unvisited,
  for (nodesIT = nodes->begin(); nodesIT != nodes->end(); nodesIT++) {
    unvisitedNodes->add(nodes->getKey(nodesIT));
  }
  
  // as long as there are unvisited nodes in the CG,
  // update the CG
  while (! unvisitedNodes->isEmpty()) {

    id = unvisitedNodes->pop(); // select and remove an unvisited node ID
    currNode = nodes->getNode(id); // get the node via ID

    // iterate through all avaialbe transtions
    for (tsIT=transitions->begin(); tsIT != transitions->end(); tsIT++) {
      
      // save refrence to transition
      currTrans = transitions->getTransition(tsIT);
      
      // check if transtion has already been fired on this node
      // and if this transition can be fired on the node
      if ((!currNode->hasApplied(transitions->getID(tsIT))) && 
	  (currTrans->applicable(currNode))) {
	
	// apply transition to the node
	newNode = currTrans->apply(currNode);
	
	// get all smaller equal predecessor nodes in the CG
	preNodes = nodes->getSmallerPredNodesIds(newNode,id);
	
	// using smaller pred. nodes, build covering node (omega-places)
	for (idIT=preNodes->begin(); idIT != preNodes->end(); idIT++) {
	  nodes->getNode(preNodes->get(idIT))->buildCoveringNode(newNode);
	}
	
	// add this node to the node set, if there is an equal node
	// already in the CG, only its ID will be returned,
	// and no new node inserted.
	newId = nodes->add(newNode);
	
	// update father/predecessor information
	nodes->getNode(newId)->setFather(id,k);
	
	// stating, that Transition k has been applied to currNode
	
	nodes->getNode(newId)->addAppliedTransitions(k);
	//currNode->addAppliedTransitions(k); 
	
	nodes->getNode(newId)->addPre(id); // currNode is a predecessor of newNode
	
	visitedNodes->add(id);// Node with given id is visited now
	
	// visit new generated node only, if node is unvisited
	if((! visitedNodes->has(newId)) &&
	   (! unvisitedNodes->has(newId))) { 
	  unvisitedNodes->add(newId);// add newId to be visited
	}
      }
    }
  }
  
  // delete temporary variables
  delete visitedNodes;
  delete preNodes;

  currNode = NULL;
  newNode = NULL;
  delete currNode;
  delete newNode;
  
  currTrans = NULL;
  delete currTrans;

  // print and return Result
  if (unvisitedNodes->isEmpty()) {
    if (!this->quiet) {
      cout << "Building the Coverability Graph is finished\n";
      cout << this->asString();
      cout << "Time elapsed: " << this->getTime() << " sec.\n";
    } 
    return true;
  }
  return false;
}  

// check if any of given markings is covered by this CG
bool CG::covers(NodeSet* testNodes) {
  for (testNodesIT = testNodes->begin(); 
       testNodesIT != testNodes->end(); testNodesIT++) {
    for (nodesIT = nodes->begin(); nodesIT != nodes->end(); nodesIT++) {
      if (*(testNodes->getNode(testNodesIT)) <= nodes->getNode(nodesIT)) {
	return true;
      }
    }
  }
  return false;
}

// combine run- and covers-method from above.
bool CG::run_covers(NodeSet* testNodes) {
  if(this->run()) {
    if (this->covers(testNodes)) {
      if(this->quiet) {
	cout << "true\n";
      } else {
	cout << "Coverability Graph covers test marking(s)\n";
      }
      return true;
    } else {
      if(this->quiet) {
	cout << "false\n";
      } else {
	cout << "Coverability Graph does not cover test marking(s).\n";
      }         
      return false;
    }
  } else {
    cout << "An Error occured while building the Coverability Graph!\n";
    std::exit(1);
  }
  return false;
}

// two places will be unioned to one,
// marks will be saved in place1, place2 will be deleted
void CG::unionPlaces(string place1, string place2) {
  this->unionPlaces(placeNames->getInt(place1),
		    placeNames->getInt(place2));
}

// two places will be unioned to one,
// marks will be saved in place1, place2 will be deleted
void CG::unionPlaces(int place1, int place2) {
  nodes->unionPlaces(place1,place2);
  transitions->unionPlaces(place1,place2);
  placeNames->del(place2);
}
