#include "InputCG.h"
#include <fstream>
#include <cstdlib>
#include <sstream>

/* constructor: initalize variables */
InputCG::InputCG() { 
  marking = new Node;
  testMarkings = new NodeSet;
  tSet = new TransitionSet;
  placeNames = new stringMap;
  transitionNames = new stringMap;
  verb = false;
}

/* provide access to parsed information:
   transitions, markings and names of places and transitions 
*/
TransitionSet* InputCG::getTransitionSet() { return tSet; }
Node* InputCG::getMarking() { return marking; }
NodeSet* InputCG::getTestMarkings() { return testMarkings; }
stringMap* InputCG::getTransitionNames() { return transitionNames; }
stringMap* InputCG::getPlaceNames() { return placeNames; }

// indicate to print more messages to console
void InputCG::setVerbose() { verb = true; }

// return string-represenation of parsed transitions and markings
string InputCG::asString() {
  ostringstream os;
  os << "Initial Marking: "
     << this->getMarking()->asString(placeNames)
     << "Test Markings\n"
     << this->getTestMarkings()->asString(placeNames)
     << "Net:\n"
     << this->getTransitionSet()->asString(placeNames,transitionNames);
  return os.str();
}

/* read the contents of a text file,
   and return content as a string,
   skipping blanks.
*/
string InputCG::readFile(const char* file) { 
  ostringstream os;
  char ch;
  std::ifstream input(file);
  if (!input) { 
    this->error_exit("Failed to open File: \""+string(file)+"\".",true);
  }
  
  while (input.get(ch)) { 
    if ((ch != ' ') && (ch != '\n')) {
      os.put(ch);
    }
  }
  return os.str();
}

/*
  parse input files for information in simple LoLA-format.
  file1 has places, initial marking and Transition net,
  file2 has end/test markings.
  If any error occures, return false,
  else return true.
*/
bool InputCG::run(const char* file1,const char* file2) { 
  string s;
  string::size_type a,b,c; // needed for String-Parsing
  string token = "";
  string error_str = ""; // used for often used error messages
  
  s = this->readFile(file1);

  error_str = "Error while parsing Places in File \""+ string(file1) + "\":\n";
  
  // get the place names
  token = "PLACE";
  a = s.find(token);
  if(a < s.size()) {
    b = s.find(";");
    if(b < s.size()) {
      this->setPlaces(string(s,a,(b-a)+1));
      s.erase(0,b+1);
    } else {
      this->error_exit(error_str + "No delimiting \";\" found",true);
      return false;
    }
  } else {
    this->error_exit(error_str + "Token \"PLACE\" not found.",true);
    return false;
  }
  
  error_str = "Error while parsing Marking in File \""+ string(file1) + "\":\n";
  
  token = "MARKING";

  // get initial marking
  a = s.find(token);
  if(a < s.size()) {
    b = s.find(";");
    if(b < s.size()) {
      this->marking = this->parseMarking(string(s,a,(b-a)+1),this->placeNames);
      s.erase(0,b+1);
    } else {
      this->error_exit(error_str + "No delimiting \";\" found",true);
    }
  } else {
    this->error_exit(error_str +"Token \"MARKING\" not found.",true);
    return false;
  }
  
  error_str = "Error while parsing Transition in File \""+ string(file1) + "\":\n";
  
  // parse Transitions
  token = "TRANSITION";
  a = s.find(token);
  while((!(s.empty())) && a < s.size()) {
    b = s.find(";");
    c = s.find(";",b+1);
    if((b > a) && (c > a) && (c < s.size()) && (b < c)) {
      this->parseTransition(string(s,a,(c-a)+1), this->tSet,
			    this->transitionNames,this->placeNames,true);
      s.erase(0,c+1);
    } else {
      this->error_exit(error_str + "No matching \";\" found",true);
      return false;
    }
    a = s.find(token);
  }
  
  error_str = "Error while parsing Markings in File \""+ string(file2) + "\":\n";
  
  s = this->readFile(file2);
  
  // Parse 2nd File, more than 1 end-marking possible
  for (a = s.find("MARKING"); a < s.size(); a = s.find("MARKING")) {
    b = s.find(";");
    if (b < s.size()) {
      testMarkings->add(this->parseMarking(string(s,a,(b-a)+1),this->placeNames));
      s.erase(0,b+1);
    } else {
      this->error_exit(error_str + "No delimiting \";\" found",true);
      return false;
    }
  }
  
  if(verb) { cout << this->asString(); }
  
  return true; // parsing was OK
}

// prints out given string, and exits acc. to input
void InputCG::error_exit(const string in, bool exit = false) {
  cout << in << endl; 
  if (exit) { std::exit(1); }
}

// parse place names given in input-string
void InputCG::setPlaces(string in) {
  const string token = "PLACE";
  string::size_type c;
  string place;
  in.erase(0,token.size());
  c = in.find_first_of(",;");
  
  while(c < in.size()) {
    place = string(in,0,c);
    if (!(place.empty())) { 
      placeNames->add(place);      
    }
    in.erase(0,c+1);
    c = in.find_first_of(",;");
  }
}

/*
  Parse Transition-String in LoLA-Format,
  and add to Transition Set.
  Expects strings of the form: 
  "TRANSITION name CONSUME place1 : c_val1, ..., placeN : c_valN;
  PRODUCE place1 : p_val1, ..., placeN : p_valN;"
  Possible Errors:
  (1) Marking refers to unknown place or place name is empty
  (2) Value for some place is not an integer
*/
bool InputCG::parseTransition(string in,TransitionSet* tSet,
			      stringMap* transitions,stringMap* places,
			      bool exit = false) {
  Transition* newT = new Transition; // used to save parsed place-value-pairs
  string::size_type a,b,c;
  string place,val,trans;
  int place_int, val_int,transition_int;
  int res_code = 0;
  string error_str = ""; // used for error messages
  string token = "TRANSITION";
  error_str = "Error while parsing Transition:\n"; 
  
  // begin: parse Transition token
  b = in.find(token);
  if (b > in.size()) { // "TRANSITION" not found
    this->error_exit(error_str+"Token \"TRANSITTION\" not found.",
		     exit);
    return false;
  }
  in.erase(0,b+token.size());
  
  // begin: parse Transition name
  token = "CONSUME";
  b = in.find(token);
  if (b > in.size()) { // name delimter not found
    this->error_exit(error_str+"Token \"CONSUME\" not found.",
		     exit);
    return false;
  }
  
  trans = string(in,0,b); // name of Transition
  if (trans.empty()) {
    this->error_exit(error_str+"Empty Transition name found.",exit);
    return false;
  }
  
  error_str = "Error while parsing Transition \""+trans+"\":\n";
  
  transition_int = transitions->has(trans); // Transition Name already available?
  if (transition_int > 0) {
    this->error_exit(error_str + "Transition Name already available.",exit);
    return false;
  }
  
  in.erase(0,b+token.size());
  // end: parse Transition name
    
  // begin: parse CONSUME-Funktion
  token = "PRODUCE";
  a = in.find(token); 
  if (a >= in.size()) { // delimiter not found
    this->error_exit(error_str+"Token \"PRODUCE\" not found.",
		     exit);
    return false;
  }
  
  c = in.find_first_of(",;");  // find token delimter for place-value-pair
  while(c < a) { // "," or ";" found before "PRODUDE"
    b = in.find(":");
    if (b < c) {  // ":" found, check for place name and value
      
      place = string(in,0,b); // place name
      if (place.empty()) { 
	
	this->error_exit(error_str + "Empty place name found.",
			 exit);
	return false;
      }
      place_int = places->has(place);
      if (place_int == 0) {
	this->error_exit(error_str + 
			 "Transition refers to unknown place \"" + place + "\".",
			 exit);
	return false;
      } else { // place name is OK, check its value
	val = string(in,b+1,c-b); // place value
	res_code = sscanf(val.c_str(),"%d",&val_int);
	if (res_code == 0) {
	  this->error_exit(error_str + "Value at place \"" +
			   place + "\" is not an Integer value.",
			   exit);
	  return false;
	} else {
	  newT->setIn(place_int,val_int);
	}
      }
    }
    in.erase(0,c+1);
    a = in.find(token); 
    c = in.find_first_of(",;");  // find token delimter for place-value-pair
  }
  
  // end: parse CONSUME-Funktion

  token = "PRODUCE";
  a = in.find(token); 
  in.erase(0,a+token.size()); // remove "PRODUCE" token
  
  // begin: parse PRODUCE-Funktion
  c = in.find_first_of(",;");  // find token delimter for place-value-pair
  while((!(in.empty())) && c < in.size()) { // "," or ";" found
    b = in.find(":");
    if (b < c) {  // ":" found, check for place name and value
      place = string(in,0,b); // place name
      if (place.empty()) { 
	this->error_exit(error_str + "Empty place name found.",
			 exit);
	return false;
      }
      place_int = places->has(place);
      if (place_int == 0) {
	this->error_exit(error_str + 
			 "Transition refers to unknown place \"" + place + "\".",
			 exit);
	return false;
      } else { // place name is OK, check its value
	val = string(in,b+1,c-b); // place value
	res_code = sscanf(val.c_str(),"%d",&val_int);
	if (res_code == 0) {
	  this->error_exit(error_str + "Value at place \"" +
			   place + "\" is not an Integer value.",
			   exit);
	  return false;
	} else {
	  newT->setOut(place_int,val_int);
	}
      }
    }
    in.erase(0,c+1);
    c = in.find_first_of(",;");  // find token delimter for place-value-pair
  }
  // end: parse PRODUCE-Funktion
  
  // get ID of new transition after adding to Transition Set,
  // and save this ID in Transition Names
  if (newT->empty()) {
    this->error_exit(error_str + "Transition is empty.",true);
    return false;
  } else {
    transitions->set(trans,tSet->add(newT)); // add transition to transition set
    return true;
  }
  return false;
}

/* 
   Expects strings of the form: 
   "MARKING ... ;"]
   and returns a new Instance of Class Node.
   possible Errors:
   (1) Marking refers to unknown place or place name is empty
   (2) Value for some place is not an integer
*/
Node* InputCG::parseMarking(string in, stringMap* places) {
  const string token = "MARKING";
  int resCode = 0;
  string error_str = "Error while parsing Marking:\n";
  Node* newNode = new Node; // used to save parsed marking information
  string::size_type b,c;
  string place,val;
  int place_int, val_int;
  in.erase(0,token.size());
  
  c = in.find_first_of(",;");
  while((!(in.empty())) && c < in.size()) {
    b = in.find(":");
    
    if (b < c) {
      place = string(in,0,b);
      val = string(in,b+1,c-b);
      
      // check if place-name is not empty
      if (!place.empty()) {
	place_int = places->has(place);
	
	if (place_int == 0) { // error, place name unknown
	  this->error_exit(error_str + 
			   "Marking refers to unknown Place \""+
			   place+"\"",true);
	} else { // place name known, parse Integer value
	  resCode = sscanf(val.c_str(),"%d",&val_int);
	  if (resCode == 0) { // error, could not parse Integer
	    this->error_exit(error_str + 
			     "No Integer value at place \""+
			     place+"\"",true);
	  } else {
	    newNode->set(place_int,val_int); // update node
	  }
	}
      } else {
	this->error_exit(error_str + 
			 "Marking includes empty place name."+
			 place+"\"",true);
      }
    }
    in.erase(0,c+1); // erase sub-string until one of [,;]
    c = in.find_first_of(",;"); // find next delimiter
  }
  return newNode; 
}
