/* Main program (interactive) for the coverability graph (CG) algorithm.
   2 Files in LoLA-Format are needed, 
   the first having the transition-net and initial marking,
   the second file having the final- or test-marking.
*/

#include "InputCG.h"
#include "CG.h"

InputCG* in = new InputCG;
CG* cg;

// print usage information
void usage(const char* name) {
  cerr << "Interactive Version of CG Program\n"
       <<"Usage: " << name << " [ -v | -q ] File1 File2\n"
       << "Option -v:\tverbose mode\n"
       << "Option -q:\tquiet mode\n"
       << "File1:\t\tInitial Marking and P/N-Net in LoLA format\n"
       << "File2:\t\tTest Marking(s) in LoLA format\n";
  cerr.flush();
  std::exit(1);
}

// menu, asking the user for changes to make
int menu() {
  int res_code = 0;
  int res = 0;
  string input;

  cout << "====================== MENU ==========================\n"
       << "Please choose from Menu by entering Number and Return:\n"
       << "[1] Add a new Place\n"
       << "[2] Union two Places\n"
       << "[3] Add new Transition\n"
       << "[4] Show place names\n"
       << "[5] Show transition names\n"
       << "[6] Exit\n"
       << "====================== MENU ==========================\n"
       << "Please enter menu number: ";
  cin >> input;
  cin.clear();
  res_code = sscanf(input.c_str(),"%d",&res);
  while (res_code == 0) {
    cout << "Error: Please enter menu number: ";
    cin >> input;
    cin.clear();
    res_code = sscanf(input.c_str(),"%d",&res);
  }
  cin.clear();
  cout << endl;
  cout.flush();
  std::system("clear");
  return res;
}

// read in a whole line, and return its contents as string, 
// having removed blanks
string read() {
  string str = "";
  string::size_type index;
  
  // seems to be a bug: using getline twice, but doing only one
  getline(cin,str);
  getline(cin,str);
 
  // replace blanks
  for (index = str.find(" "); index < str.size(); index = str.find(" ")) {
    str.replace(index,1,"");
  }
  return str;
}

// user can add a new place to already set of known places.
void addPlace() {
  string newPlace;
  cout << "Please enter the name of the new Place (without whitespace): ";
  cin.clear();
  cin >> newPlace;
  cin.clear();
  cout << endl;
  cout.flush();
  if (cg->getPlaceNames()->has(newPlace)) {
    cout << "Place Name: \"" << newPlace << "\" already available!\n";
  } else {
    cg->getPlaceNames()->add(newPlace);
    if(cg->getPlaceNames()->has(newPlace)) {
      cout << "Place Name: \"" << newPlace << "\" is inserted now.\n";
    } else {
      cout << "An error occured while trying to insert Place Name: \"" 
	   << newPlace 
	   << "\", please try again.\n";
    }
  }
}

// user can union two places to one,
// values are stored under the name of the first given place,
// the second 
void unionPlaces() {
  string place1,place2;
  
  cin.clear();
  cout << "Please enter the name of the first Place: ";
  cin >> place1;
  cin.clear();
  cout << "Please enter the name of the second Place: ";
  cin >> place2;
  cin.clear();
  if (cg->getPlaceNames()->has(place1) &&
      cg->getPlaceNames()->has(place2)) {
    in->getTestMarkings()->unionPlaces(cg->getPlaceNames()->getInt(place1),
				       cg->getPlaceNames()->getInt(place2));
    cg->unionPlaces(place1,place2);
    
    cout << "Places \""
	 << place1 << "\" and \""
	 << place2 
	 << "\" are unioned now in place \""
	 << place1 << "\"" << endl;
    
    cg->run_covers(in->getTestMarkings());
  } else {
    cout << "Error: At least one of the places is not available.\n";
  }
}

// add a new transition to already available set of transitions
void addTransition() {
  string input = "";
  cout << "Please enter the new Transition in LoLA-Format and press Return:\n";
  input = read();
  if (in->parseTransition(input,
			  cg->getTransitionSet(),
			  cg->getTransitionNames(),
			  cg->getPlaceNames(),
			  false)) {
    cg->run_covers(in->getTestMarkings());
    cout << "New Transition inserted and re-build Coverability Graph.\n";
  } else {
    cout << "New Transition not inserted due to error.\n";
  }
}

/* present a menu, and loop until user signals exit,
   asking user for updates and re-building the CG
*/
int main(int argc, char* argv[]) {
  bool verb = false; // verbose messages  
  bool quiet = false; // quiet mode
  
  char* file1; // save file names
  char* file2;
  string verbose = "-v";
  string quiet_str = "-q";

  bool loop = true;
  
  if (argc < 3) { usage(argv[0]); }
  
  if (argv[1] == verbose) {
    verb = true;
    file1 = argv[2];
    file2 = argv[3];
  } else {
    if (argv[1] == quiet_str) {
      quiet = true;
      file1 = argv[2];
      file2 = argv[3];
    } else {
      file1 = argv[1];
      file2 = argv[2];
    }
  }
  
  if (verb) { in->setVerbose(); }
  
  if (in->run(file1,file2)) {
    cg = new CG(in->getTransitionSet(),
		in->getMarking(),
		in->getPlaceNames(),
		in->getTransitionNames());
    
    if(verb) { cg->setVerbose(); }
    if (quiet) { cg->setQuiet(); }
    
    cg->run_covers(in->getTestMarkings());
    
    while(loop) {
      cin.clear();
      cout.clear();
      switch(menu()) {
      case 1: // add a new/unknown place
	addPlace();
	break;
      case 2: // union 2 places
	unionPlaces();
	break;
      case 3: // add a new transition
	addTransition();
	break;
      case 4: // print available places
	cout << "Place names = " << cg->getPlaceNames()->asString();
	break;
      case 5: // print available transitions
	cout << "Transition names = " << cg->getTransitionNames()->asString();
	break;
      case 6: // exit
	loop = false;
	std::exit(0);
	break;
      default:
	break;
      }
    }
  } else { // error in input-files
    cout << "Error: Could not read Petri-Net description and Test Markings from ";
    cout << file1 << " and " << file2 << endl;
    std::exit(1);
  }
}
