/**
 * pg2lola.cpp
 * Transforms the output of the aunfold tool to the lola format.
 * The result is written to standard out.
 */

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include <list>
#include <map>
#include <stdexcept>
#include <iostream>

using namespace std;

/**
 * A place id with its initial marking
 */
typedef pair<string,int> placetype;

/**
 * The list of all place id's
 */
list<placetype> places;

/**
 * Maps place id's to labels
 */
map<string,string> placeLabels;

/**
 * The list of all transition id's
 */
list<string> transitions;

/**
 * Maps transition id's to labels
 */
map <string,string> transitionLabels;

/**
 * A place id and the number of tokens
 * produced for or consumed from that place
 */
 typedef pair<string,int> transitionItem;

/**
 * Maps transitions to transitionItems. An entry tells how many tokens
 * are consumed from that place.
 */
map<string,list<transitionItem> > consume;

/**
 * Maps transitions to transitionItems. An entry tells how many tokens
 * are produced for that place.
 */
map<string,list<transitionItem> > produce;

/**
 * returns the value of the user attribute (ATTR) of the node with the specified name of the
 * specified type or NULL if noch such attribute exists
 */
const char* getUserAttrValue(xmlNodePtr cur, const char* name, const char* type)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }

  if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
  {
    cur = cur->xmlChildrenNode;
    while ( (cur != NULL) && xmlStrcasecmp((const xmlChar*) type, cur->name) )
    {
      cur = cur->next;
    }
    if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
    {
      return (const char*) cur->xmlChildrenNode->content;
    }
  }
  return NULL;
}

/**
 * returns true if the given node has an user attribute (ATTR) with the specified name,
 * false otherwise
 */
bool hasUserAttr(xmlNodePtr cur, const char* name)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }
  return (cur != NULL);
}

/**
 * processes the rel cur is pointing
 */
void processPetrinetRel(xmlNodePtr cur)
{
  const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
  if (id == NULL)
  {
    throw runtime_error("Rel id expected");
  }
  const char* initial_marking = getUserAttrValue(cur, "initial_marking", "int");
  if (initial_marking == NULL)
  {
    initial_marking = "0";
  }
  const char* label = getUserAttrValue(cur, "label", "string");
  if (label == NULL)
  {
    label = "";
  }

  //store the place with its initial marking
  places.push_back(placetype(id, atoi(initial_marking)));

  //remove blanks and store the label of the place
  string s = label;
  while (s.find(" ") != string::npos)
    s.replace(s.find(" "), 1, "");
  placeLabels[id] = s;

  cur = cur->xmlChildrenNode;
  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "relend"))
    {
      const xmlChar* role = xmlGetProp(cur, (const xmlChar*) "role");
      if ( (!xmlStrcasecmp(role, (const xmlChar*) "preset")) ||
           (!xmlStrcasecmp(role, (const xmlChar*) "postset")) )
      {
        const char* target = (const char*) xmlGetProp(cur, (const xmlChar*) "target");
        const char* weight = getUserAttrValue(cur, "weight", "int");
        if (target == NULL)
        {
          throw runtime_error("Relend target expected");
        }
        if (weight == NULL)
        {
          weight = "1";
        }
        int w = atoi(weight);
        if (!xmlStrcasecmp(role, (const xmlChar*) "preset"))
        {
          consume[target].push_back(transitionItem(id, w));
        }
        else
        {
          produce[target].push_back(transitionItem(id, w));
        }
      }
    }
    cur = cur->next;
  }
}

/**
 * extracts the petrinet out of the petrigraph cur is pointing to
 */
void extractPetrinet(xmlNodePtr cur)
{
  //step down to the first petrigraph element
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "node"))
    {
      //we are only interested in nodes with the attribute 'transition'
      if (hasUserAttr(cur, "transition"))
      {
        const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
        if (id != NULL)
        {
          const char* rule = getUserAttrValue(cur, "rule", "string");
          if (rule == NULL)
          {
            rule = "";
          }
          //store the transition
          transitions.push_back(id);

          //remove blanks and store the label of the transition
          string s = rule;
          while (s.find(" ") != string::npos)
            s.replace(s.find(" "), 1, "");
          transitionLabels[id] = s;
        }
        else
        {
          throw runtime_error("Transition id expected");
        }
      }
    }
    else if (!xmlStrcasecmp(cur->name, (const xmlChar*) "rel"))
    {
      processPetrinetRel(cur);
    }
    //other elements (neither node nor rel) are ignored

    cur = cur->next;
  }
}

/**
 * prints usage information
 */
void Usage()
{
  cout << "\nUsage: pg2lola filename\n\n";
}

int main(int argc, char **argv)
{

  try
  {
    xmlDocPtr doc;
    xmlNodePtr cur;

    //process command line arguments
    if (argc < 2)
    {
      Usage();
      return -1;
    }

    //parse the input file and obtain an xmlDocPtr
    doc = xmlParseFile(argv[1]);
    if (doc == NULL)
    {
      throw runtime_error("XML parser error");
    }

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL)
    {
      throw runtime_error("Empty xml document");
    }
    if (xmlStrcasecmp(cur->name, (const xmlChar*) "gxl"))
    {
      throw runtime_error("Root node 'GXL' expected");
    }

    cur = cur->xmlChildrenNode;
    if (cur == NULL)
    {
      throw runtime_error("Graph expected");
    }

    while ( (cur != NULL) && xmlStrcasecmp(cur->name, (const xmlChar*) "graph") )
    {
      cur = cur->next;
    }

    if (cur == NULL)
    {
      throw runtime_error("Graph expected");
    }

    extractPetrinet(cur);

    //places
    cout << "PLACE ";
    list<placetype>::iterator it;
    for (it = places.begin(); it != places.end(); it++)
    {
      if (it != places.begin())
      {
        cout << ", ";
      }
      cout << placeLabels[it->first] << it->first;
    }
    cout << ";\n\n";

    //markings
    cout << "MARKING ";
    bool firstEntry = true;
    list<placetype>::iterator it2;
    for (it2 = places.begin(); it2 != places.end(); it2++)
    {
      int m = it2->second;
      if (m != 0)
      {
        if (! firstEntry)
        {
          cout << ", ";
        }
        cout << placeLabels[it2->first] << it2->first << ": " << m;
        firstEntry = false;
      }
    }
    cout << ";\n\n";

    //transitions
    list<string>::iterator it3;
    for (it3 = transitions.begin(); it3 != transitions.end(); it3++)
    {
      string transition = *it3;
      cout << "TRANSITION " << transitionLabels[transition] << transition << "\n";

      //consume
      cout << "CONSUME ";
      map<string,list<transitionItem> >::iterator it4 = consume.find(transition);
      if (it4 != consume.end())
      {
        list<transitionItem> consumeItems = it4->second;
        list<transitionItem>::iterator it5;
        for (it5 = consumeItems.begin(); it5 != consumeItems.end(); it5++)
        {
          if (it5 != consumeItems.begin())
          {
            cout << ", ";
          }
          cout << placeLabels[it5->first] << it5->first << ": " << it5->second;
        }
      }
      cout << ";\n";

      //produce
      cout << "PRODUCE ";
      map<string,list<transitionItem> >::iterator it6 = produce.find(transition);
      if (it6 != produce.end())
      {
        list<transitionItem> produceItems = it6->second;
        list<transitionItem>::iterator it7;
        for (it7 = produceItems.begin(); it7 != produceItems.end(); it7++)
        {
          if (it7 != produceItems.begin())
          {
            cout << ", ";
          }
          cout << placeLabels[it7->first] << it7->first << ": " << it7->second;
        }
      }
      cout << ";\n\n";
    }

    return 0;
  }
  catch (exception& e)
  {
    cout << "*** " << e.what() << " ***\n";
  }
}
