/**
 * pg2neato.cpp
 * Transforms the output of the aunfold tool to the neato format.
 * The result is written to standard out.
 */

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include <map>
#include <math.h>
#include <stdexcept>
#include <stdio.h>
#include <string>
#include <iostream>

using namespace std;

/**
 * Common neato attributes for hypergraph nodes (comma seperated list)
 */
string HG_NODE_ATTRIBUTES = "shape=circle, width=.15, label=\"\"";


/**
 * Common neato attributes for hyperedges (comma seperated list)
 */
string HYPEREDGE_ATTRIBUTES = "shape=box";

/**
 * Common neato attributes for petrinet transitions (comma seperated list)
 */
string PN_TRANSITION_ATTRIBUTES = "shape=box";

/**
 * Common neato attributes for petrinet places (comma seperated list)
 */
string PN_PLACE_ATTRIBUTES = "shape=ellipse";

/**
 * returns the value of the user attribute (ATTR) of the node with the specified name of the
 * specified type or NULL if noch such attribute exists
 */
const char* getUserAttrValue(xmlNodePtr cur, const char* name, const char* type)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }

  if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
  {
    cur = cur->xmlChildrenNode;
    while ( (cur != NULL) && xmlStrcasecmp((const xmlChar*) type, cur->name) )
    {
      cur = cur->next;
    }
    if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
    {
      return (const char*) cur->xmlChildrenNode->content;
    }
  }
  return NULL;
}

/**
 * returns true if the given node has an user attribute (ATTR) with the specified name,
 * false otherwise
 */
bool hasUserAttr(xmlNodePtr cur, const char* name)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }
  return (cur != NULL);
}

/**
 * processes the rel cur is pointing to for a hypergraph output
 */
void processHypergraphRel(xmlNodePtr cur)
{
  const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
  if (id == NULL)
  {
    throw runtime_error("Rel id expected");
  }
  const char* label = getUserAttrValue(cur, "label", "string");
  if (label == NULL)
  {
    label = "";
  }
  cout << id << "[label=\"" << label << "\\n" << id << "\", " << HYPEREDGE_ATTRIBUTES << "]\n";

  cur = cur->xmlChildrenNode;
  map<string, int> edgeLabelPos;
  while (cur != NULL)
  {
    if ( (!xmlStrcasecmp(cur->name, (const xmlChar*) "relend")) &&
         (!xmlStrcasecmp((const xmlChar*) "vertex", xmlGetProp(cur, (const xmlChar*) "role"))) )
    {
      const char* target = (const char*) xmlGetProp(cur, (const xmlChar*) "target");
      if (target == NULL)
      {
        throw runtime_error("Relend target expected");
      }
      const char* startorder = (const char*) xmlGetProp(cur, (const xmlChar*) "startorder");
      if (startorder == NULL)
      {
        throw runtime_error("Relend startorder expected");
      }
      int pos = edgeLabelPos[target] + 1;
      float angle = 0;
      int dist = 0;
      const float PI = 3.1415;
      if ((pos % 2)  == 0)
      {
        angle = atan(tan(35.0*PI/180)/(pos/2))/PI*180;
        dist = pos / 2;
      }
      else
      {
        angle = - atan(tan(35.0*PI/180)/((pos+1)/2))/PI*180;
        dist = (pos+1)/2;
      }
      cout << id << " -- " << target << "[taillabel=\"" << startorder;
      cout << "\", labelangle=" << angle << ", labeldistance=" << dist << "]\n";
      edgeLabelPos[target]++;
    }
    cur = cur->next;
  }
}

/**
 * processes the rel cur is pointing to for a petrinet output
 */
void processPetrinetRel(xmlNodePtr cur)
{
  const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
  if (id == NULL)
  {
    throw runtime_error("Rel id expected");
  }
  const char* label = getUserAttrValue(cur, "label", "string");
  if (label == NULL)
  {
    label = "";
  }
  cout << id << " [label=\"" << label << "\\n" << id << "\", " << PN_PLACE_ATTRIBUTES;
  cout << "]\n";

  cur = cur->xmlChildrenNode;
  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "relend"))
    {
      const xmlChar* role = xmlGetProp(cur, (const xmlChar*) "role");
      if ( (!xmlStrcasecmp(role, (const xmlChar*) "preset")) ||
           (!xmlStrcasecmp(role, (const xmlChar*) "postset")) )
      {
        const char* target = (const char*) xmlGetProp(cur, (const xmlChar*) "target");
        const char* weight = getUserAttrValue(cur, "weight", "int");
        if (target == NULL)
        {
          throw runtime_error("Relend target expected");
        }
        if (weight == NULL)
        {
          weight = "";
        }
        if (!xmlStrcasecmp(role, (const xmlChar*) "preset"))
        {
          cout << id << " -- " << target << "[dir=forward, taillabel=\"";
          cout << weight << "\"]\n";
        }
        else
        {
          cout << target << " -- " << id << "[dir=forward, taillabel=\"";
          cout << weight << "\"]\n";
        }
      }
    }
    cur = cur->next;
  }
}

/**
 * extracts the hypergraph out of the petrigraph cur is pointing to
 */
void extractHypergraph(xmlNodePtr cur)
{
  //step down to the first petrigraph element
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "node"))
    {
      //we are only interested in nodes with the attribute 'vertex'
      if (hasUserAttr(cur, "vertex"))
      {
        const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
        if (id != NULL)
        {
          cout << id << " [" << HG_NODE_ATTRIBUTES << "]\n";
        }
        else
        {
          throw runtime_error("Node id expected");
        }
      }
    }
    else if (!xmlStrcasecmp(cur->name, (const xmlChar*) "rel"))
    {
      processHypergraphRel(cur);
    }
    //other elements (neither node nor rel) are ignored

    cur = cur->next;
  }
}

/**
 * extracts the petrinet out of the petrigraph cur is pointing to
 */
void extractPetrinet(xmlNodePtr cur)
{
  //step down to the first petrigraph element
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "node"))
    {
      //we are only interested in nodes with the attribute 'transition'
      if (hasUserAttr(cur, "transition"))
      {
        const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
        const char* rule = getUserAttrValue(cur, "rule", "string");
        if (id == NULL)
        {
          throw runtime_error("Transition id expected");
        }
        if (rule != NULL)
        {
          cout << id << " [label=\"" << rule << "\\n" << id;
          cout << "\", " << PN_TRANSITION_ATTRIBUTES << "]\n";
        }
        else
        {
          cout << id << " [label=\"" << id << "\", " << PN_TRANSITION_ATTRIBUTES;
          cout << "]\n";
        }
      }
    }
    else if (!xmlStrcasecmp(cur->name, (const xmlChar*) "rel"))
    {
      processPetrinetRel(cur);
    }
    //other elements (neither node nor rel) are ignored

    cur = cur->next;
  }
}

/**
 * prints usage information
 */
void Usage()
{
  cout << "\nUsage: pg2neato (-pn|-hg) filename\n";
  cout << "       -pn     output petri net\n";
  cout << "       -hg     output hypergraph\n\n";
}

/**
 * output type: hypergraph or petrinet
 */
enum modetype {HG, PETRI};

int main(int argc, char **argv)
{

  try
  {
    modetype mode;
    xmlDocPtr doc;
    xmlNodePtr cur;

    //process command line arguments
    if (argc < 3)
    {
      Usage();
      return -1;
    }
    if (!strcmp(argv[1], "-pn"))
    {
      mode = PETRI;
    }
    else if (!strcmp(argv[1], "-hg"))
    {
      mode = HG;
    }
    else
    {
      Usage();
      return -1;
    }

    string neatoTitle = argv[2];
    int pos = neatoTitle.rfind('/');
    if (pos != -1)
    {
      neatoTitle = neatoTitle.substr(pos + 1, neatoTitle.length());
    }
    pos = neatoTitle.rfind(".xml");
    if (pos != -1)
    {
      neatoTitle = neatoTitle.substr(0, pos);
    }

    //parse the input file and obtain an xmlDocPtr
    doc = xmlParseFile(argv[2]);
    if (doc == NULL)
    {
      throw runtime_error("XML parser error");
    }

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL)
    {
      throw runtime_error("Empty xml document");
    }
    if (xmlStrcasecmp(cur->name, (const xmlChar*) "gxl"))
    {
      throw runtime_error("Root node 'GXL' expected");
    }

    cur = cur->xmlChildrenNode;
    if (cur == NULL)
    {
      throw runtime_error("Graph expected");
    }

    while ( (cur != NULL) && xmlStrcasecmp(cur->name, (const xmlChar*) "graph") )
    {
      cur = cur->next;
    }

    if (cur == NULL)
    {
      throw runtime_error("Graph expected");
    }

    cout << "graph \"" << neatoTitle << "\" {\n";

    if (mode == HG)
    {
      extractHypergraph(cur);
    }
    else
    {
      extractPetrinet(cur);
    }

    cout << "}\n";
    return 0;
  }
  catch (exception& e)
  {
    cout << "*** " << e.what() << " ***\n";
  }
}
