/**
 * pg2pnml.cpp
 * Transforms the output of the aunfold tool to the pnml format
 * without layout information.
 */

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include <list>
#include <map>
#include <stdexcept>
#include <stdio.h>

using namespace std;

/**
 * A place id with its initial marking
 */
typedef pair<string,int> placetype;

/**
 * The list of all place id's
 */
list<placetype> places;

/**
 * Maps place id's to labels
 */
map<string,string> placeLabels;

/**
 * The list of all transition id's
 */
list<string> transitions;

/**
 * Maps transition id's to labels
 */
map <string,string> transitionLabels;

/**
 * An arc consists of source id, target id and inscription
 */
typedef pair<pair<string,string>,int> arctype;

/**
 * The list of all arcs
 */
list <arctype> arcs;

/**
 * returns the value of the user attribute (ATTR) of the node with the specified name of the
 * specified type or NULL if noch such attribute exists
 */
const char* getUserAttrValue(xmlNodePtr cur, const char* name, const char* type)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }

  if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
  {
    cur = cur->xmlChildrenNode;
    while ( (cur != NULL) && xmlStrcasecmp((const xmlChar*) type, cur->name) )
    {
      cur = cur->next;
    }
    if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
    {
      return (const char*) cur->xmlChildrenNode->content;
    }
  }
  return NULL;
}

/**
 * returns true if the given node has an user attribute (ATTR) with the specified name,
 * false otherwise
 */
bool hasUserAttr(xmlNodePtr cur, const char* name)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }
  return (cur != NULL);
}

/**
 * processes the rel cur is pointing
 */
void processPetrinetRel(xmlNodePtr cur)
{
  const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
  if (id == NULL)
  {
    throw runtime_error("Rel id expected");
  }
  const char* initial_marking = getUserAttrValue(cur, "initial_marking", "int");
  if (initial_marking == NULL)
  {
    initial_marking = "0";
  }
  const char* label = getUserAttrValue(cur, "label", "string");
  if (label == NULL)
  {
    label = "";
  }

  //store the place with its initial marking
  places.push_back(placetype(id, atoi(initial_marking)));

  //store the label of the place
  placeLabels[id] = label;


  cur = cur->xmlChildrenNode;
  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "relend"))
    {
      const xmlChar* role = xmlGetProp(cur, (const xmlChar*) "role");
      if ( (!xmlStrcasecmp(role, (const xmlChar*) "preset")) ||
           (!xmlStrcasecmp(role, (const xmlChar*) "postset")) )
      {
        const char* target = (const char*) xmlGetProp(cur, (const xmlChar*) "target");
        const char* weight = getUserAttrValue(cur, "weight", "int");
        if (target == NULL)
        {
          throw runtime_error("Relend target expected");
        }
        if (weight == NULL)
        {
          weight = "1";
        }
        int w = atoi(weight);
        if (!xmlStrcasecmp(role, (const xmlChar*) "preset"))
        {
          arcs.push_back(arctype(pair<string,string>(id, target), w));
        }
        else
        {
          arcs.push_back(arctype(pair<string,string>(target, id), w));
        }
      }
    }
    cur = cur->next;
  }
}

/**
 * extracts the petrinet out of the petrigraph cur is pointing to
 */
void extractPetrinet(xmlNodePtr cur)
{
  //step down to the first petrigraph element
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "node"))
    {
      //we are only interested in nodes with the attribute 'transition'
      if (hasUserAttr(cur, "transition"))
      {
        const char* id = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
        if (id != NULL)
        {
          const char* rule = getUserAttrValue(cur, "rule", "string");
          if (rule == NULL)
          {
            rule = "";
          }
          //store the transition
          transitions.push_back(id);

          //store the label of the transition
          transitionLabels[id] = rule;
        }
        else
        {
          throw runtime_error("Transition id expected");
        }
      }
    }
    else if (!xmlStrcasecmp(cur->name, (const xmlChar*) "rel"))
    {
      processPetrinetRel(cur);
    }
    //other elements (neither node nor rel) are ignored

    cur = cur->next;
  }
}

/**
 * prints usage information
 */
void Usage()
{
  printf("\nUsage: pg2pnml infile outfile\n\n");
}

int main(int argc, char **argv)
{

  try
  {
    xmlDocPtr docIn, docOut;
    xmlNodePtr cur;

    //process command line arguments
    if (argc < 3)
    {
      Usage();
      return -1;
    }

    //parse the input file and obtain an xmlDocPtr
    docIn = xmlParseFile(argv[1]);
    if (docIn == NULL)
    {
      throw runtime_error("XML parser error");
    }

    cur = xmlDocGetRootElement(docIn);
    if (cur == NULL)
    {
      throw runtime_error("Empty xml document");
    }
    if (xmlStrcasecmp(cur->name, (const xmlChar*) "gxl"))
    {
      throw runtime_error("Root node 'GXL' expected");
    }

    cur = cur->xmlChildrenNode;
    if (cur == NULL)
    {
      throw runtime_error("Graph expected");
    }

    while ( (cur != NULL) && xmlStrcasecmp(cur->name, (const xmlChar*) "graph") )
    {
      cur = cur->next;
    }

    if (cur == NULL)
    {
      throw runtime_error("Graph expected");
    }

    extractPetrinet(cur);

    docOut = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL, (const xmlChar *)"pnml");
    xmlDocSetRootElement(docOut, root);

    xmlNodePtr ptNet = xmlNewChild(root, NULL, (const xmlChar *)"net", NULL);
    xmlSetProp(ptNet, (const xmlChar *)"type", (const xmlChar *) "PTNet");
    xmlSetProp(ptNet, (const xmlChar *)"id", (const xmlChar *)"n1");

    string ptNetTitle = argv[1];
    int pos = ptNetTitle.rfind('/');
    if (pos != -1)
    {
      ptNetTitle = ptNetTitle.substr(pos + 1, ptNetTitle.length());
    }
    pos = ptNetTitle.rfind(".xml");
    if (pos != -1)
    {
      ptNetTitle = ptNetTitle.substr(0, pos);
    }

    xmlNodePtr ptNetName = xmlNewChild(ptNet, NULL, (const xmlChar *)"name", NULL);
    xmlNodePtr ptNetNameValue = xmlNewChild(ptNetName, NULL, (const xmlChar *)"value", (const xmlChar*) ptNetTitle.data());

    //places
    list<placetype>::iterator it;
    for (it = places.begin(); it != places.end(); it++)
    {
      xmlNodePtr place = xmlNewChild(ptNet, NULL, (const xmlChar *)"place", NULL);
      xmlSetProp(place, (const xmlChar *)"id", (const xmlChar *) it->first.data());

      xmlNodePtr placeName = xmlNewChild(place, NULL, (const xmlChar *)"name", NULL);
      xmlNodePtr ptNetNameValue = xmlNewChild(placeName, NULL, (const xmlChar *)"value", (const xmlChar*) placeLabels[it->first.data()].data());

      char buf[256];
      sprintf(buf, "%d", it->second);
      xmlNodePtr marking = xmlNewChild(place, NULL, (const xmlChar *)"initialMarking", NULL);
      xmlNodePtr markingValue = xmlNewChild(marking, NULL, (const xmlChar *)"value", (const xmlChar*) buf);
    }

    //transitions

    list<string>::iterator it3;
    for (it3 = transitions.begin(); it3 != transitions.end(); it3++)
    {
      xmlNodePtr transition = xmlNewChild(ptNet, NULL, (const xmlChar *)"transition", NULL);
      xmlSetProp(transition, (const xmlChar *)"id", (const xmlChar *) it3->data());

      xmlNodePtr transName = xmlNewChild(transition, NULL, (const xmlChar *)"name", NULL);
      xmlNodePtr ptNetNameValue = xmlNewChild(transName, NULL, (const xmlChar *)"value", (const xmlChar*) transitionLabels[it3->data()].data());
    }

    //arcs
    list<arctype>::iterator it4;
    for (it4 = arcs.begin(); it4 != arcs.end(); it4++)
    {
      const char* source = it4->first.first.data();
      const char* target =  it4->first.second.data();
      int inscription = it4->second;

      xmlNodePtr arc = xmlNewChild(ptNet, NULL, (const xmlChar *)"arc", NULL);
      xmlSetProp(arc, (const xmlChar *)"source", (const xmlChar *) source);
      xmlSetProp(arc, (const xmlChar *)"target", (const xmlChar *) target);

      xmlNodePtr inscr = xmlNewChild(arc, NULL, (const xmlChar *)"inscription", NULL);
      char buf[256];
      sprintf(buf, "%d", inscription);
      xmlNodePtr inscrValue = xmlNewChild(inscr, NULL, (const xmlChar *)"value", (const xmlChar*) buf);
    }

    if(xmlSaveFormatFile(argv[2], docOut, true) == -1)
    {
      throw runtime_error("Error writing XML file");
    }

    return 0;
  }
  catch (exception& e)
  {
    printf("*** %s ***\n", e.what());
  }
}
